import click

from qrtransfert.qr_encode import encode as qr_encode
from qrtransfert.qr_decode import decode as qr_decode

from qrtransfert.constants import DEFAULT_THRESHOLD, DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT, DEFAULT_FPS


@click.group()
def cli():
  pass

@cli.command()
@click.option("-f", "file", required=True, type=str, help="Path of a movie file to decode")
@click.option("-ct", "--correction-threshold", default=DEFAULT_THRESHOLD, help="Image correction threshold", type=click.IntRange(0, 255, clamp=True))
@click.option("-fps", "--fps-to-decode", default=DEFAULT_FPS, help="Desired fps for decode QRCode", type=click.IntRange(1, 60, clamp=True))
@click.option("-d", "--debug", is_flag=True, help="Only display QRCode to adjust the threshold variable")
def decode(file, correction_threshold, fps_to_decode, debug):
  """ Decode a video file with QRCodes info a single file """
  qr_decode(file, correction_threshold, fps_to_decode, debug)

@cli.command()
@click.option("-f", "file", required=True, type=click.File(mode='rb'), help="Path of a single file to transfert")
@click.option("-c", "--qr-color", is_flag=True, help="Use color QRCode to increase by 3 the data content in a QRCode")
@click.option("-t", "--terminal", is_flag=True, help="Display QRCode in terminal")
@click.option("-qmax", "--qr-max-version", default=20, help="Maximun version of the QRCode to generate", type=click.IntRange(1, 40, clamp=True))
@click.option("-qcor", "--qr-correction", default="M", help="Define the error correction level for the QRCode", type=click.Choice(['L', 'M', 'Q', 'H'], case_sensitive=False))
@click.option("-l", "--loop-delay", default=0.6, help="Set the loop rotation delay for the QRCodes in seconds", type=float)
@click.option("-ww", "--window-width", default=DEFAULT_WINDOW_WIDTH, help="Width of the windows to display QRCodes", type=int)
@click.option("-wh", "--window-height", default=DEFAULT_WINDOW_HEIGHT, help="Height of the windows to display QRCodes", type=int)
@click.option("-o", "--out-to-file", help="Output filename", type=str)
def encode(file, qr_color, terminal, qr_max_version, qr_correction, loop_delay, window_width, window_height, out_to_file):
  """ Transfert a single file using QRCodes """
  qr_encode(file, qr_color, terminal, qr_max_version, qr_correction, loop_delay, window_width, window_height, out_to_file)

if __name__ == "__main__":
  cli()


import cv2
import sys
import json
import time
import base64
import hashlib
import numpy as np
import pyzbar.pyzbar as pyzbar

from qrtransfert.constants import DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT

def decode(file, correction_threshold, fps_to_decode, debug):
   
    cap = cv2.VideoCapture(file)

    if not cap.isOpened():
        print(f"Erreur : Impossible to open file '{file}'")
        return []

    original_fps = cap.get(cv2.CAP_PROP_FPS)
    if original_fps == 0:
        print("Warning: Unable to get video framerate. Processing all frames.")
        frames_to_skip = 0
    else:
        frames_to_skip = int(original_fps / fps_to_decode)
        if frames_to_skip < 1:
            frames_to_skip = 1
        print(f"Original framerate of the video : {original_fps:.2f} FPS")
        print(f"Processing at approximately {fps_to_decode} FPS. Will skip {frames_to_skip - 1} frame between each processing")

    decoded_data_set = set()
    decoded_data_list = None
    filename = None
    src_checksum = None
    frame_count = 0

    debug_windows_name = "Debug - Press 'Q' to exit"
    cv2.namedWindow(debug_windows_name, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(debug_windows_name, DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT)

    while True:
        ret, frame = cap.read()

        if not ret:
            break

        frame_count += 1

        if frame_count % frames_to_skip == 0 or frames_to_skip == 0:
            
            rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

            r_frame = np.zeros_like(rgb_frame)
            r_only = rgb_frame[:, :, 0]
            r_frame[:, :, 0] = r_only
            r_frame[:, :, 1] = r_only
            r_frame[:, :, 2] = r_only
            r_frame = cv2.cvtColor(r_frame, cv2.COLOR_BGR2GRAY)

            g_frame = np.zeros_like(rgb_frame)
            g_only = rgb_frame[:, :, 1]
            g_frame[:, :, 0] = g_only
            g_frame[:, :, 1] = g_only
            g_frame[:, :, 2] = g_only
            g_frame = cv2.cvtColor(g_frame, cv2.COLOR_BGR2GRAY)

            b_frame = np.zeros_like(rgb_frame)
            b_only = rgb_frame[:, :, 2]
            b_frame[:, :, 0] = b_only
            b_frame[:, :, 1] = b_only
            b_frame[:, :, 2] = b_only
            b_frame = cv2.cvtColor(b_frame, cv2.COLOR_BGR2GRAY)

            for f in (r_frame, g_frame, b_frame):
                # Clean image
                _, clean_frame = cv2.threshold(f, correction_threshold, 255, cv2.THRESH_BINARY)

                if debug:
                    resized_clean_frame = cv2.resize(clean_frame, (DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT), interpolation=cv2.INTER_AREA)
                    cv2.imshow(debug_windows_name, resized_clean_frame)
                    time.sleep(0.3)
                    if cv2.waitKey(10) & 0xFF == ord('q'):
                        sys.exit(0)
                else:
                    decoded_objects = pyzbar.decode(clean_frame, symbols=[pyzbar.ZBarSymbol.QRCODE])
                    for obj in decoded_objects:

                        sequence_number = int.from_bytes(obj.data[0:2], byteorder='big')
                        total_sequences = int.from_bytes(obj.data[2:4], byteorder='big')

                        # Initialize array with good size
                        if decoded_data_list is None:
                            decoded_data_list = [None] * (total_sequences -1)

                        # Speedup process
                        if sequence_number != total_sequences:
                            if decoded_data_list[int(sequence_number-1)] is not None:
                                continue
                        else:
                            if filename is not None:
                                continue

                        if None not in decoded_data_list and filename is not None:
                            break

                        print(f"Got part: {sequence_number}/{total_sequences}")

                        decoded_data = obj.data[4:].decode('utf-8')
                        
                        if decoded_data not in decoded_data_set:
                            decoded_data_set.add(decoded_data)

                            if sequence_number == total_sequences:
                                decoded_data = base64.b64decode(decoded_data)
                                json_data = json.loads(decoded_data)
                                filename = json_data['filename'] if json_data['filename'] else "noname"
                                src_checksum = json_data['checksum'] if json_data['checksum'] else "000"
                            else:
                                decoded_data_list[(sequence_number-1)] = decoded_data
    
    if not debug:
        cap.release()
        cv2.destroyAllWindows()

        if decoded_data_list:
            nb_parts = sum(1 for item in decoded_data_list if item is not None)
            nb_missing_parts = int(total_sequences-nb_parts-1)
            if nb_missing_parts > 0:
                print(f"Missing some part of the file ({nb_missing_parts})")
            else:
                base64_string = ''.join(decoded_data_list)
                binary_data = base64.b64decode(base64_string)
                checksum = hashlib.md5(binary_data).hexdigest()
                with open(filename, 'wb') as f:
                    f.write(binary_data)
                print(f"File '{filename}' successfully wrote")
                if checksum != src_checksum:
                    print("WARNING: the checksum is not the same that the original file")




import os
import cv2
import json
import time
import base64
import hashlib
import keyboard
import numpy as np
import tkinter as tk
import matplotlib.pyplot as plt

from tqdm import tqdm
from time import sleep
from PIL import Image, ImageTk
from qrcode import QRCode, constants

from qrtransfert.utils import max_bytes_per_version


def chunk_byte(data, length):
  return [data[i:i + length] for i in range(0, len(data), length)]

def generate_qr_code(data, sequence_number, total_sequences):

    error_correction_map = {
        "L": constants.ERROR_CORRECT_L,
        "M": constants.ERROR_CORRECT_M,
        "Q": constants.ERROR_CORRECT_Q,
        "H": constants.ERROR_CORRECT_H
    }

    data = int(sequence_number).to_bytes(2, 'big') + int(total_sequences).to_bytes(2, 'big') + data

    # Set max verion or auto when there is less data
    version_to_use=QR_MAX_VERSION
    if len(data) < MAX_QR_CODE_SIZE:
        version_to_use=None

    qr_code = QRCode(
        version = version_to_use,
        error_correction = error_correction_map[QR_CORRECTION]
    )
    qr_code.add_data(data, optimize=0)
    qr_code.make()
    qr_code_image = qr_code.make_image(fill_color="black", back_color="white")

    qr_code_image = qr_code_image.resize((QR_GEN_SIZE, QR_GEN_SIZE)).convert("L")
    
    return qr_code_image

def on_press_q(key):
    keyboard.unhook_all()
    stop_loop()

def stop_loop():
    global STOP_LOOP
    STOP_LOOP=True

def clear_console():
    os.system('cls' if os.name=='nt' else 'clear')

def display_qrcodes_in_terminal(qrcodes):

    # Listen for press 'q'
    keyboard.on_press_key("q", on_press_q)

    image_index = -1
    while True:
        if STOP_LOOP:
            clear_console()
            break
        clear_console()
        image_index = (image_index + 1) % len(qrcodes)
        qrcode = qrcodes[image_index]
        plt.imshow(qrcode)
        plt.axis('off')
        plt.show(block=False)
        print("Press Q to exit")
        sleep(LOOP_DELAY)
        plt.close()

def display_qrcodes(qrcodes):

    # Listen for press 'q'
    keyboard.on_press_key("q", on_press_q)

    def display_next_image():

        qrcode = qrcodes[image_index]
        image = qrcode
        image = image.resize((WINDOW_WIDTH, WINDOW_HEIGHT))
        photo = ImageTk.PhotoImage(image)

        label_index = tk.Label(window, text=f"Image {image_index + 1}/{len(qrcodes)}", font=("Arial", 12))
        label_index.place(relx=0.5, y=(WINDOW_HEIGHT+25), anchor=tk.CENTER)

        label_image.configure(image=photo)
        label_image.image = photo

    window = tk.Tk()
    window.geometry(WINDOW_SIZE)
    frame_image = tk.Frame(window)
    frame_image.pack()
    label_image = tk.Label(frame_image)
    label_image.pack()

    # Listen for press 'q'
    window.protocol("WM_DELETE_WINDOW", stop_loop)

    image_index = -1
    print("Press Q to exit")
    while True:
        if STOP_LOOP:
            window.destroy()
            break
        image_index = (image_index + 1) % len(qrcodes)
        display_next_image()
        window.update_idletasks()
        window.update()
        sleep(LOOP_DELAY)

def encode(file, qr_color, terminal, qr_max_version, qr_correction, loop_delay, window_width, window_height, out_to_file):

    # Update global variables
    global QR_MODE
    QR_MODE="byte"
    global BYTE_CONTROL
    BYTE_CONTROL=4
    global STOP_LOOP
    STOP_LOOP=False
    global LOOP_DELAY
    LOOP_DELAY=loop_delay
    global QR_CORRECTION
    QR_CORRECTION=qr_correction
    global QR_MAX_VERSION
    QR_MAX_VERSION=qr_max_version
    global MAX_QR_CODE_SIZE
    MAX_QR_CODE_SIZE=max_bytes_per_version(qr_max_version, qr_correction, QR_MODE)
    global MAX_SIZE
    MAX_SIZE=(MAX_QR_CODE_SIZE - BYTE_CONTROL)
    global WINDOW_WIDTH
    WINDOW_WIDTH=window_width
    global WINDOW_HEIGHT
    WINDOW_HEIGHT=window_height
    global WINDOW_SIZE
    WINDOW_SIZE=f"{WINDOW_WIDTH}x{(WINDOW_HEIGHT+50)}"
    global QR_GEN_SIZE
    QR_GEN_SIZE=1050

    # Def vars
    total_chunks = 2
    chunks = []
    qrcodes = []

    # Convert data to base64
    filename = os.path.basename(file.name)
    file_content = file.read()
    checksum = hashlib.md5(file_content).hexdigest()
    binary_data = base64.b64encode(file_content)

    if len(binary_data) > MAX_SIZE:
        chunks = chunk_byte(binary_data, MAX_SIZE)
    else:
        chunks = [binary_data]

    total_chunks = (len(chunks) + 1)

    # Estimated minimum transfer time
    ett=round(loop_delay*(total_chunks/3))
    m, s = divmod(ett, 60)
    h, m = divmod(m, 60)
    print(f'Minimum transfer time: {h:d}h {m:02d}m {s:02d}s')

    # Generate QRCodes
    with tqdm(len(chunks)) as pbar:
        for i, chunk in enumerate(chunks):
            pbar.set_description(f"Generating QRCode {i+1}/{total_chunks}")
            qrcode = generate_qr_code(chunk, (i + 1), total_chunks)
            qrcodes.append(qrcode)

        # Generate last QRCode
        pbar.set_description(f"Generating QRCode {total_chunks}/{total_chunks}")
        json_data = {
            "filename": f"{filename}",
            "checksum": f"{checksum}"
        }
        json_string = json.dumps(json_data)
        last_data = json_string.encode("utf-8")
        last_data_base64 = base64.b64encode(last_data)
        qrcode = generate_qr_code(last_data_base64, total_chunks, total_chunks)
        qrcodes.append(qrcode)

    qr_to_display = qrcodes

    if qr_color:
        # Merge to RGB
        nb_bands = 3
        splited_qrcodes = [qrcodes[i:i+nb_bands] for i in range(0, len(qrcodes), nb_bands)]
        empty_image = Image.new('RGB', (1, 1), (255, 255, 255)).resize((QR_GEN_SIZE, QR_GEN_SIZE)).convert("L")
        chroma_qrcodes = []
        for i in splited_qrcodes:
            while len(i) != nb_bands:
                i.append(empty_image)
            chroma_qrcodes.append(Image.merge('RGB', i))

        qr_to_display = chroma_qrcodes

    # Display QRCodes
    if out_to_file:
        fps = 10
        fourcc = cv2.VideoWriter.fourcc(*'MJPG')
        video = cv2.VideoWriter(out_to_file, fourcc, fps, (QR_GEN_SIZE,QR_GEN_SIZE))
        for image in qr_to_display:
            cv_img = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
            for _ in range(fps*2):
                video.write(np.array(cv_img))
        cv2.destroyAllWindows()
        video.release()
    else:
        if terminal:
            display_qrcodes_in_terminal(qr_to_display)
        else:
            display_qrcodes(qr_to_display)
